#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

/*
 * First some definitions for frequently-used macros, blatently
 * "borrowed" from Georgia Tech CS2130 defs.h file.
 */

/*
 * The ERR_MSG macro prints a message on stderr describing where an
 * error occured and what its cause was (if known)
 */

#define ERR_MSG( func )                                         \
    {                                                           \
        ( void )fflush( stderr );                               \
        ( void )fprintf( stderr, __FILE__ ":%d:" #func "\n",\
                __LINE__ );                  \
    }

/*
 * PRINTF_LIKE macro
 *
 * This macro is a wrapper for functions (like printf(3s))
 * that people tend to want to ignore the return value of.
 * Since printf(3s) can have a variable number of arguments, we
 * must rely on a "trick" to get the macro properly expanded.
 *
 * To use this macro, substitute PRINTF(( ... )) everywhere you
 * would normally use printf( ... ). Note the use of the double
 * parentheses (that is the "trick").
 */

#define PRINTF_LIKE( fn, args, test )   if( fn args test ) ERR_MSG( fn )

/*
 * PRINTF_LIKE functions - these require the (( ... )) "trick"
 * because they take a variable number of arguments
 */

#define PRINTF( args )          PRINTF_LIKE( printf, args, < 0 )
#define FPRINTF( args )         PRINTF_LIKE( fprintf, args, < 0 )
#define SPRINTF( args )         PRINTF_LIKE( sprintf, args, < 0 )

#define SCANF( args )           PRINTF_LIKE( scanf, args, < 0 )
#define FSCANF( args )          PRINTF_LIKE( fscanf, args, < 0 )
#define SSCANF( args )          PRINTF_LIKE( sscanf, args, < 0 )

/*
 * Define a function to calculate integer powers.
 * This allows me not to have to compile with -lm
 */

static double Pow(double base, int expon);
static double Pow_helper(double base, int expon)
{
   return (expon==0)?1:base*Pow(base,expon-1);
}
static double Pow(double base, int expon)
{
   if(expon < 0)
      return 1.0 / Pow_helper(base, (-expon));
   else
      return Pow_helper(base, expon);
}

/* Define size of a string of user input */
#define STRSIZE 2048

/* Define Enumerated types */
enum { DONE, OK, DOEVAL, NUMBER, OPERATOR, LPAREN, RPAREN, QUIT, ERROR };
/* Define math operators */
#define OPERATORS "+-*/~^"
/* Global Variable, yes I know it's bad */
int PARENCOUNT = 0;

/*
 * Gets input from keyboard.  Forces something to be inputted 
 * Modifies input string
 */
void get_input( char *input )
{
   PRINTF(( "> " ));

   /* Extracts input from keyboard */
   while( fgets( input, STRSIZE, stdin ) == NULL ) 
   {
      PRINTF(( "> " ));
   }
}

/* 
 * Does the standard Math funcs.
 * Result is stored in first value.
 */
void do_math_operations( double *first, double second, char op )
{
   switch( op )
   {
      case '+':                        /* Add         */
         *first = *first + second;
         break;
      case '-':                        /* Subtract    */
         *first = *first - second;
         break;
      case '*':                        /* Multiply    */
         *first = *first * second;
         break;
      case '/':                        /* Divide      */
         *first = (double)*first / second;
         break;
      case '^':                        /* Do Power Of */
         *first = Pow( *first, (int)second );
         break;
      case 'x':                        /* set first to val2 */
         *first = second;
         break;
      default :
         PRINTF(( "Eureka! You have a serious problem...\n" ));
         PRINTF(( "Aborting...\n" ));
         exit( EXIT_FAILURE );
         break;
   }
}

/*
 * Recognizes next token from input stream.
 * Returns enumerated type describing token
 * Stores token in string in memory pointed to by next_input
 * Pushes user_input pointer to next token for next call.
 */
static int recognize_next( char **user_input, char *next_input )
{
   int 
      num_left;

   /* Scan next token */
   if(( num_left = sscanf( *user_input, "%s", next_input )) <= 0 )
      return DONE;

   /* Push pointer to next token */
   *user_input = strchr( *user_input, next_input[0] );
   *user_input += strlen( next_input );

   /* Figure out what we have here */
   if( isdigit( next_input[0] ) || isdigit( next_input[1] ))
   {
      return NUMBER;
   }
   else if( strchr( OPERATORS, next_input[0] ) != NULL )
   {
      return OPERATOR;
   }
   else switch( toupper(next_input[0]) )
   {
	  case 'Q':
         return QUIT;
		 break;
      case '(':
         return LPAREN;
		 break;
      case ')':
         return RPAREN;
		 break;
      default:
         return DONE;
		 break;
   }
}

/* Recursive module to calculate a "stack" of data
 * input is a pointer to text input, and gets moved 
 * around.  value holds the "answer" provided by each call.
 */
int do_math( char **input, double *value )
{
   char
      next[ STRSIZE ],          /* Stores user input            */
      op='x';                   /* Operator                     */
   int
      status = OK,              /* enum for operation status    */
      type;                     /* type of input extracted      */
   double
      val2;                     /* Values read into program     */

   if( input == NULL || *input == NULL )
   {
      return ERROR;
   }

   /* We expect to see a number first or LPAREN or quit*/
   if(( type = recognize_next( input, next )) != NUMBER  &&
        type != QUIT && type != RPAREN && type != LPAREN )
   {
      return ERROR;
   }
   else switch( type )
   {
      /* If we see an LPAREN, call self for parens reduction */
	  case LPAREN:
	     PARENCOUNT++;
         status = do_math( input, value );
         /* If we have pulled FIRST value, we want to pull another */ 
	     if( op == 'x' )
		    status = OK;
	     break;
      /* if we see RPAREN, we know syntax is wrong */
      case RPAREN: 
         PARENCOUNT--;
         return ERROR; 
         break;
      case QUIT: 
         return QUIT;
		 break;
      default: /* Process value like normal */
         *value = atof( next );
		 break;
   }

   /* Expect to see operator then number  or DONE*/ 
   while( status != ERROR && status != DONE )
   {
      if(( type = recognize_next( input, next )) != OPERATOR &&
           type != DONE && type != RPAREN && type != LPAREN )
      {  
         return ERROR;
      }
      /* If we see an LPAREN, call self for parens reduction */
      else switch( type )
      {
		 case LPAREN:
		    PARENCOUNT++;
            status = do_math( input, &val2 );
			break;
         /* if we see RPAREN here we're done with parens */
         case RPAREN: 
            PARENCOUNT--;
            return DOEVAL; 
            break;
         case DONE:
            return DONE;
			break;
         default: /* process operator */
            op = next[0];
			break;
      }

      if( status != DOEVAL && /* Don't want to pull another value if
							   * we reduced a parens */
		  (type = recognize_next( input, next )) != NUMBER &&
           type != RPAREN && type != LPAREN ) 
      {  
         return ERROR;
      }
      else switch( type )
      {  
 	     /* Bad syntax if find RPAREN here */
		 case RPAREN:
		    PARENCOUNT--;
            return ERROR; 
			break;
         /* If we see an LPAREN, call self for parens reduction */
         case LPAREN: 
            PARENCOUNT++;
            status = do_math( input, &val2 );
			break;
         default:
			if( status != DOEVAL )
               val2 = atof( next );
            break;
      }

      /* If everything is ok, process math ops */
      if( status != ERROR && status != DONE ) 
      {
         do_math_operations( value, val2, op ); 
		 status = OK; /* Set status to OK so we can continue */
	  }
   } 

   return status;
}

int main( void )
{
   char 
      *input,      
      *ptr;
   double
      Answer = 0;
   int 
      status = OK;

   /* While user still wants to process data, keep going */
   while( status != QUIT )
   {
      PARENCOUNT = 0;
	  if(( input = malloc( STRSIZE * sizeof( char ) )) == NULL )
      {
         FPRINTF(( stderr, "Malloc error! Aborting...\n" ));
         exit( EXIT_FAILURE );
      }

	  /* get user input, store in area pointed to by input */
      get_input( input );
	  ptr = input;

	  /* Run the program and see how we did */
	  status = do_math( &ptr, &Answer );
	  /* Check for even parens */
	  if( PARENCOUNT != 0 )
		  status = ERROR;
      switch(status)
      {
         case DONE:
		 case OK:
            PRINTF(( "ans = %2.5f\n", Answer ));
            break;
         case QUIT:
            break;
         case ERROR:
            PRINTF(( "ERROR: Expression Invalid\n" ));
            break;
         default:
            PRINTF(( "Bad input from Calculator, aborting...\n" ));
            exit( EXIT_FAILURE );
      }
      
	  /* Wipe out memory area for next user input */
      free( input );
   }
   PRINTF(( "Thank You, Come Again!\n" ));

   return( EXIT_SUCCESS ); 
}

